#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <stdarg.h>

/* ----------------------------------------------------------------- *
 * Written by Feline (feline@irendi.com) - Irendi Software
 *
 * All code distributed under the GNU Public Licence, version 2.
 * Copyright Michael Sharpe 1994-2000.
 *
 * We grant permission to use, copy, modify, and distribute this software
 * under the terms of the GPL v2, provided that the above copyright
 * notice and this text are not removed. We make no guarantee about the
 * sutability of this software for any purpose and we are not liable
 * for any damages resulting from its use. Further, we are under no
 * obligation to maintain or extend this software. It is provided on an
 * "as is" basis without any express or implied warranty.
 *
 * Bug reports, enhancement requests and bug fixes are always welcome.
 *
 * History 
 * v1.0 March 94 -- initial revision
 * v1.1 Feb 95 -- improve randomness
 * v2.0 13/Oct/96 -- remove need for shell script which did the grep -c
 *                   Allow a user defined string to be used to seperate quotes.
 *                   Added -h option :) well not really.
 *                   Purified - Fixed some UMR's
 *                              Fixed some fd leaks
 * v2.1 12/Jan/97 -- Daemon version. Probably does not work everywhere now.
 *                   I tested it on Solaris 2.5.1 and HP-UX
 * v2.2 16/Nov/99 -- major clean up.
 *                   remove all the Mother() random number rubbish
 *                   fix pid file to allow for multiple users
 *                   first public release
 * v2.3 19/May/01 -- detect is the autosig daemon is already running
 *                   Yen-Ming Lee <leeym@bsd.ce.ntu.edu.tw>
 *
 *  Notes:
 *  No line should be more than 255 characters long!!
 * ----------------------------------------------------------------- */

#define MAX_LINE_LENGTH 256

char *g_programName;

static void 
gratuitousPlug() {
   printf("Auto-sig v2.3\n"
          "Written by Feline (feline@irendi.com) - Irendi Software\n"
          "Credits - Yen-Ming Lee (leeym@bsd.ce.ntu.edu.tw) - detect if autosig is running\n"
          "Auto-sig invoked as '%s'\n\n", g_programName);
}

static void 
usage(void) {
   printf("Usage :-\n");
   printf("%s quoteFilename delimiter header sig_file [daemon] [create]\n\n",
          g_programName);
   printf("quoteFile = file of *quotes*\n");
   printf("delimiter = characters which delimit quotes\n");
   printf("   header = file contents will form the top of the sig file.\n");
   printf("  sigFile = file to change, .plan or .sig say.\n");
   printf("   daemon = daemon (run in daemon mode).\n");
}

#define E_WARN 1
#define E_ERROR 2
#define E_INFO 3
static void
pOutput(int type, char *fmt, ...) {
   va_list ap;
   FILE *fp;

   switch (type) {
      case E_WARN: fprintf(stderr, "WARNING: "); fp = stderr; break;
      case E_ERROR: fprintf(stderr, "ERROR: "); fp = stderr; break;
      case E_INFO: fprintf(stdout, "INFO: "); fp = stdout; break;
      default:
         break;
   }

   va_start(ap, fmt);
   vfprintf(fp, fmt, ap);
   va_end(ap);
}


static int
openNamedPipe(char *name, int mode, mode_t perms) {
   mode_t old_umask;
   int fd; 
   int status;
   struct stat statBuf;

   status = stat(name, &statBuf);
   if (status == -1) {
      if (errno == ENOENT) {
         /* the named pipe does not exist, so create it now */
         old_umask = umask(022);
         status = mkfifo(name,perms);
         (void)umask(old_umask);
      } else {
         pOutput(E_ERROR, "Could not stat named pipe `%s'\n", name);
      }
   } else {
      if (! S_ISFIFO(statBuf.st_mode)) {
         pOutput(E_ERROR, "`%s' is not a fifo\n", name);
         status = -1;
      }
   }

   if (status != -1)
      fd = open(name, mode, perms);
   else
      fd = -1;

   return fd;
}

static int
countQuotesInFile(char *name,char *delimiter) {
    char line[MAX_LINE_LENGTH];
    int delimiterSize = strlen(delimiter);
    FILE *fp;
    int count = -1;
 
    fp = fopen(name, "r");
    if (fp) {
       while (!feof(fp)) {
          fgets(line, MAX_LINE_LENGTH, fp); 
          if (feof(fp)) 
             break;
          if ( 0 == strncmp(line, delimiter, delimiterSize)) 
             count++;
       }
       fclose(fp);
    }
    return count;
}

static int
getRandomNumber(int range) {
    static int seeded = 0;
    time_t s1;
    pid_t s2;
    int seed;
    int num;

    if (!seeded) {
       s1 = time(NULL);
       s2 = getpid();
       seed = s1 / s2;
       srand(seed);
       seeded = 1;
    }
    num = rand();
    num = num % range + 1;

    return num;
}

static int
writeHeader(int fd, char *headerFilename) {
   FILE *fp;
   char line[MAX_LINE_LENGTH];

   fp = fopen(headerFilename, "r");
   if (fp) {
      while (!feof(fp)) {
         line[0] = '\0';
         fgets(line, MAX_LINE_LENGTH, fp);
         write(fd, (void *)line, strlen(line));
      }
      fclose(fp);
   }
   return fp ? 1 : 0;
}

static int
writeQuoteFromFile(int fd, char *quoteFilename, char *delimiter, int num) {
   FILE *fp;
   char line[MAX_LINE_LENGTH];
   int count = 0;
   int delimiter_size = strlen(delimiter);

   fp = fopen(quoteFilename, "r");
   if (fp) {
      while (!feof(fp)) {
         fgets(line, MAX_LINE_LENGTH, fp); 
         if ( 0 == strncmp(line, delimiter, delimiter_size)) 
            count++;
         if (count == num) break;
      }
      while (!feof(fp)) {
         line[0] = '\0';
         fgets(line, MAX_LINE_LENGTH, fp); 
         if ( 0 == strncmp(line, delimiter, delimiter_size)) 
            break;
         write(fd, (void *)line, strlen(line));
      }
      fclose(fp);
   }
   return fp ? 1 : 0;
}

#ifdef NEED_BASENAME
static char *
basename(char *path) {
   char *p;

   p = strrchr(path, '/');
   if (p) {
      p++;
      if (*p == '\0') 
         p = path;
   } else {
      p = path;
   }
   return p;
}
#endif
  
static char *
getUserName(void) {
   char *username;
   struct passwd *pw;

   pw = getpwuid(getuid());
   if (pw)
      username = pw->pw_name;
   else
      username = "unknown";
   return username;
}

static int
constructPidFileName(char *buf, size_t bufSize) {
   char *username = getUserName();
   char *autosigName = basename(g_programName);
   int status = 0;
   size_t len;

   len = strlen(username) + strlen(autosigName) + 10 + 1;
   if (len < bufSize) {
      sprintf(buf, "/tmp/%s-%s.pid", autosigName, username);
      status = 1;
   }
   return status;
}

static pid_t
daemonIsRunning(void) {
   char *username = getUserName();
   char pidFilename[256];
   FILE *pidFp;
   int isRunning = 0;
   int status;
   pid_t pid;

   status = constructPidFileName(pidFilename, sizeof(pidFilename)); 
   if (status) {
      pidFp = fopen(pidFilename, "r");
      if (pidFp) {
         /* found a pid file which might indicate that autosig is 
          * still running */
         fscanf(pidFp, "%d", &pid);
         fclose(pidFp);
         if (pid)
            isRunning = !kill(pid, 0);
      }
   }
   return isRunning ? pid : 0;
}

/*****************************************
 * argv[0] == program name
 * argv[1] == quote file
 * argv[2] == quote delimiter
 * argv[3] == header file
 * argv[4] == sig_file
 * argv[5] == if present the program will run as a daemon writing to a named
 *            pipe.
 */
int 
main(int argc, char *argv[]) {
   int range, num;
   int sigFd;
   int status, exitStatus = 0;
   char *quoteFilename, *delimiter, *headerFilename, *outputFilename;

   g_programName = argv[0];
   
   if ((argc < 5) || (argc > 6)) {
      gratuitousPlug();
      usage();
   } else {
      quoteFilename = argv[1];
      delimiter = argv[2];
      headerFilename = argv[3];
      outputFilename = argv[4];

      if (argc < 6) {
         range = countQuotesInFile(quoteFilename, delimiter);
         if (range >= 0) {
            num = getRandomNumber(range);
            sigFd = open(outputFilename, O_WRONLY|O_CREAT|O_TRUNC, 0644);
            if (sigFd != -1) {
               status = writeHeader(sigFd, headerFilename);
               if (status) {
                  status = writeQuoteFromFile(sigFd, quoteFilename, 
                                              delimiter, num);
                  if (!status) {
                     pOutput(E_ERROR, "Could not read quoteFile `%s'\n",
                            quoteFilename);
                     exitStatus = 1;
                  }
               } else {
                  pOutput(E_ERROR, "Could not read header file `%s'\n",
                          headerFilename);
                  exitStatus = 1;
               }
               close(sigFd);
            } else {
               pOutput(E_ERROR, "Could not open output file `%s'\n",
                       outputFilename);
               exitStatus = 1;
            }
         } else {
            pOutput(E_ERROR, "Could not read quote file `%s' to count "
                             "quotes\n", quoteFilename);
            exitStatus = 1;
         }
      } else {
         /* we are running as a daemon.... */
         pid_t pid;
         pid_t runningPid;
   
         gratuitousPlug();

         runningPid = daemonIsRunning();
         if (!runningPid) {
            pid = fork();
            if (pid == -1) {
               pOutput(E_ERROR,"Could not fork child process.\n");
               exitStatus = 1;
            } else if (pid) {
               /* Parent: simply write the pid to a file in /tmp */
               FILE *pidFp;
               char pidFilename[256];
      
               pOutput(E_INFO,"Child process is '%ld'\n", (long)pid);
               constructPidFileName(pidFilename, sizeof(pidFilename));
               pidFp = fopen(pidFilename, "w");
               if (pidFp) {
                  fprintf(pidFp,"%ld\n",(long)pid);
               fclose(pidFp);
               }
               exitStatus = 0;
            } else {
               /* Child: simply loop forever. Each time through the loop, write 
                * a new quote to the named pipe 
                */
               while (1) {
                  range = countQuotesInFile(quoteFilename, delimiter);
                  if (range >= 0) {
                     sigFd = openNamedPipe(outputFilename, O_WRONLY, 0644);
                     if (sigFd != -1) {
                        num = getRandomNumber(range);
                        status = writeHeader(sigFd, headerFilename);
                        if (status) {
                           status = writeQuoteFromFile(sigFd, quoteFilename,
                                                       delimiter, num);
                           if (!status) {
                              pOutput(E_ERROR, "(child) Could not read quote"
                                              " file `%s'\n", quoteFilename);
                              exitStatus = 1;
                              break;
                           }
                        } else {
                           pOutput(E_ERROR, "(child) Could not read header "
                                           "file `%s'\n", headerFilename);
                           exitStatus = 1;
                           break;
                        }
                        close(sigFd);
                        sleep(1);
                     } else {
                        pOutput(E_ERROR, "(child) failed to open named pipe\n");
                        exitStatus = 1;
                        break;
                     }
                  } else {
                     pOutput(E_ERROR, "(child) Could not read quote file "
                                     "`%s' to count quotes\n", quoteFilename);
                     exitStatus = 1;
                     break;
                  }
               }
            }
         } else {
            pOutput(E_ERROR, "%s daemon is already running with pid %d\n",
                    basename(g_programName), runningPid);
         }
      }
   }
   return exitStatus;
}
